/*
 *
 * php_rrdtool.c
 *
 *	PHP interface to RRD Tool. (for php4/5/zend)
 *
 *
 *       Joe Miller, <joeym@ibizcorp.com>, <joeym@inficad.com>
 *          iBIZ Technology Corp,  SkyLynx / Inficad Communications
 *          2/12/2000 & 7/18/2000
 *
 *       Jeffrey Wheat <jeff@cetlink.net> - 10/01/2002
 *       - Fixed to build with php-4.2.3
 *
 *       Dan Cech <dan@ubersmith.com>
 *         Ubersmith, Inc.
 *
 *       Benny Baumann <BenBE@geshi.org> - 07/22/2009
 *       - Added open_basedir checks
 *
 * See README, INSTALL, and USAGE files for more details.
 *
 * $Id: rrdtool.c,v 1.1.1.1 2002/02/26 10:21:20 oetiker Exp $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stddef.h>

/* PHP Includes */
#include "php.h"
#include "php_logos.h"
#include "ext/standard/info.h"
#include "SAPI.h"

/* rrdtool includes */
#include "php_rrdtool.h"
#include "rrdtool_logo.h"
#include <rrd.h>

#if HAVE_RRDTOOL

/* If you declare any globals in php_rrdtool.h uncomment this:
ZEND_DECLARE_MODULE_GLOBALS(rrdtool)
 */

function_entry rrdtool_functions[] = {
	PHP_FE(rrd_clear_error,			NULL)
	PHP_FE(rrd_create,				NULL)
	PHP_FE(rrd_dump,				NULL)
	PHP_FE(rrd_error,				NULL)
	PHP_FE(rrd_fetch,				NULL)
	PHP_FE(rrd_first,				NULL)
	PHP_FE(rrd_graph,				NULL)
	PHP_FE(rrd_info,				NULL)
	PHP_FE(rrd_last,				NULL)
	PHP_FE(rrd_lastupdate,			NULL)
	PHP_FE(rrd_restore,				NULL)
	PHP_FE(rrd_tune,				NULL)
	PHP_FE(rrd_update,				NULL)
	PHP_FE(rrd_xport,				NULL)
	PHP_FE(rrdtool_info,			NULL)
	PHP_FE(rrdtool_logo_guid,		NULL)
	{NULL, NULL, NULL}
};

zend_module_entry rrdtool_module_entry = {
	STANDARD_MODULE_HEADER,
	"rrdtool",
	rrdtool_functions,
	PHP_MINIT(rrdtool),
	PHP_MSHUTDOWN(rrdtool),
	NULL,
	NULL,
	PHP_MINFO(rrdtool),
	NO_VERSION_YET,
	STANDARD_MODULE_PROPERTIES,
};

#ifdef COMPILE_DL_RRDTOOL
ZEND_GET_MODULE(rrdtool)
#endif

#define PHP_RRD_VERSION_STRING "1.1.0"

/* {{{ PHP_MINIT_FUNCTION */
PHP_MINIT_FUNCTION(rrdtool)
{
	php_register_info_logo(RRDTOOL_LOGO_GUID, "image/gif", rrdtool_logo, sizeof(rrdtool_logo));

	//Constants for rrd_dump Header style
	#define RRD_REGISTER_CONST(x)                               \
	zend_register_long_constant(                                \
		""#x, sizeof(""#x), x,                                  \
		CONST_PERSISTENT, module_number TSRMLS_CC)
	RRD_REGISTER_CONST(RRD_DUMP_XML_NONE);
	RRD_REGISTER_CONST(RRD_DUMP_XML_DTD);
	RRD_REGISTER_CONST(RRD_DUMP_XML_XSD);
	#undef RRD_REGISTER_CONST

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION */
PHP_MSHUTDOWN_FUNCTION(rrdtool)
{
	php_unregister_info_logo(RRDTOOL_LOGO_GUID);

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION */
PHP_MINFO_FUNCTION(rrdtool)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "RRDTool support", "enabled");
	php_info_print_table_row(2, "RRDTool Library Version", rrd_strversion());
	php_info_print_table_row(2, "RRDTool Extension Version", PHP_RRD_VERSION_STRING);
#ifdef COMPILE_DL_RRDTOOL
	php_info_print_table_row(2, "RRDTool Library Linkage", "Extension");
#else
	php_info_print_table_row(2, "RRDTool Library Linkage", "Bundled");
#endif
	php_info_print_table_end();
}
/* }}} */

typedef struct rrd_arg_array {
	int count;
	char **args;
	} rrd_arg_array;

rrd_arg_array *__rrd_argument_strdup(
	char *op,
	char *preceed,
	HashTable *args) {

	HashPosition pointer;
	zval **data;

	int i = 2;

	rrd_arg_array *result = (rrd_arg_array *)emalloc(sizeof(rrd_arg_array));
	if(!result)
		return NULL;

	result->count = zend_hash_num_elements(args) + 2;
	if(preceed)
		result->count++;

	result->args = (char **) emalloc(result->count * sizeof(char *));
	if(!result->args) {
		efree(result);
		return NULL;
	}

	result->args[0] = "dummy";
	result->args[1] = estrdup(op);

	if(preceed)
		result->args[i++] = estrdup(preceed);

	for(zend_hash_internal_pointer_reset_ex(args, &pointer);
		zend_hash_get_current_data_ex(args, (void**) &data, &pointer) == SUCCESS;
		zend_hash_move_forward_ex(args, &pointer)) {

		zval temp = **data;

		zval_copy_ctor(&temp);
		convert_to_string(&temp);

		result->args[i++] = estrdup(Z_STRVAL(temp));

		zval_dtor(&temp);
	}

	return result;
}

void __rrd_argument_free(
	rrd_arg_array *args) {

	if(!args)
		return;

	if(args->args) {
		int i;

		for (i = 1; i < args->count; i++)
			efree(args->args[i]);

		efree(args->args);
	}

	efree(args);
}

/* {{{ proto mixed rrd_graph(string file, array args_arr, [int argc])
	Creates a graph based on options passed via an array */
PHP_FUNCTION(rrd_graph)
{
	char *file;
	int file_len;
	long dummy;

	zval *args, **data;
	HashTable *args_arr;

	zval *p_calcpr;

	int xsize, ysize;
#if HAVE_RRD_12X
	double ymin, ymax;
#endif

	int i;

	char **calcpr;

	if ( rrd_test_error() )
		rrd_clear_error();

	if ( zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sa|l", &file, &file_len, &args, &dummy) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(file TSRMLS_CC))
		RETURN_FALSE;

	args_arr = Z_ARRVAL_P(args);

	rrd_arg_array *argv = __rrd_argument_strdup("graph", file, args_arr);
	if(!argv)
		RETURN_FALSE;

	optind = 0; opterr = 0;

#if HAVE_RRD_12X
	if ( rrd_graph(argv->count-1, &argv->args[1], &calcpr, &xsize, &ysize, NULL, &ymin, &ymax) == -1 ) {
#else
	if ( rrd_graph(argv->count-1, &argv->args[1], &calcpr, &xsize, &ysize) == -1 ) {
#endif
		RETVAL_FALSE;
	} else {
		array_init(return_value);
		add_assoc_long(return_value, "xsize", xsize);
		add_assoc_long(return_value, "ysize", ysize);

#if HAVE_RRD_12X
		add_assoc_double(return_value, "ymin", ymin);
		add_assoc_double(return_value, "ymax", ymax);
#endif

		MAKE_STD_ZVAL(p_calcpr);
		array_init(p_calcpr);

		if (calcpr) {
			for (i = 0; calcpr[i]; i++) {
				add_next_index_string(p_calcpr, calcpr[i], 1);
				free(calcpr[i]);
			}
			free(calcpr);
		}

		add_assoc_zval(return_value,"calcpr",p_calcpr);
	}

	__rrd_argument_free(argv);

	return;
}
/* }}} */

/* {{{ proto mixed rrd_xport(array args_arr)
	Creates a graph based on options passed via an array */
PHP_FUNCTION(rrd_xport)
{
#if HAVE_RRD_12X
	zval *args;
	HashTable *args_arr;

	zval *p_legend_v, *p_return_data;

	int xxsize;
	unsigned long int j = 0;
	time_t start, end, ti;
	unsigned long step, col_cnt, row_cnt;
	rrd_value_t *return_data, *ptr;
	char **legend_v;
	int i;

	if ( rrd_test_error() )
		rrd_clear_error();

	if (ZEND_NUM_ARGS() < 1)
		WRONG_PARAM_COUNT;

	if ( zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a", &args) == FAILURE )
		RETURN_FALSE;

	args_arr = Z_ARRVAL_P(args);

	rrd_arg_array *argv = __rrd_argument_strdup("xport", NULL, args_arr);
	if(!argv)
		RETURN_FALSE;

	optind = 0; opterr = 0;

	if ( rrd_xport(argv->count-1, &argv->args[1], &xxsize, &start, &end, &step, &col_cnt, &legend_v, &return_data) != -1 ) {

		row_cnt = (end - start) / step;
		ptr = return_data;

		array_init(return_value);

		add_assoc_long(return_value, "start", start);
		add_assoc_long(return_value, "end", end);
		add_assoc_long(return_value, "step", step);
		add_assoc_long(return_value, "col_cnt", col_cnt);
		add_assoc_long(return_value, "row_cnt", row_cnt);

		ALLOC_INIT_ZVAL(p_legend_v);
		array_init(p_legend_v);

		add_next_index_string(p_legend_v, "t", 1);

		if (legend_v) {
			for (j = 0; j < col_cnt; j++) {
				char *entry;
				entry = legend_v[j];
				add_next_index_string(p_legend_v, entry, 1);
				free(entry);
			}
			free(legend_v);
		}

		add_assoc_zval(return_value,"legend",p_legend_v);

		ALLOC_INIT_ZVAL(p_return_data);
		array_init(p_return_data);

		if (return_data) {
			for (ti = start + step; ti <= end; ti += step) {
				zval *subarray;

				ALLOC_INIT_ZVAL(subarray);
				array_init(subarray);

				add_next_index_double(subarray, ti);

				for (j = 0; j < col_cnt; j++) {
					rrd_value_t newval = DNAN;

					newval = *ptr;
					if (isnan(newval)) {
						add_next_index_string(subarray, "NaN", 1);
					} else {
						add_next_index_double(subarray, newval);
					}
					ptr++;
				}

				add_next_index_zval(p_return_data, subarray);
			}
			free(return_data);
		}

		add_assoc_zval(return_value,"data",p_return_data);
	} else {
		RETVAL_FALSE;
	}

	__rrd_argument_free(argv);

	return;

#else
	//Always fail for old versions!
	RETURN_FALSE;
#endif
}
/* }}} */

/* {{{ proto mixed rrd_fetch(string file, array args_arr, [int argc])
	Fetch info from an RRD file */
PHP_FUNCTION(rrd_fetch)
{
	char *file;
	int file_len;
	zval *args;
	HashTable *args_arr;
	long dummy;

	zval *p_ds_namv, *p_data;
	int i, j, argc;
	time_t start, end;
	unsigned long step, ds_cnt;
	char **ds_namv;
	rrd_value_t *return_data, *datap;

	if ( rrd_test_error() )
		rrd_clear_error();

	if ( zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sa|l", &file, &file_len, &args, &dummy) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(file TSRMLS_CC))
		RETURN_FALSE;

	args_arr = Z_ARRVAL_P(args);

	rrd_arg_array *argv = __rrd_argument_strdup("fetch", file, args_arr);
	if(!argv)
		RETURN_FALSE;

	optind = 0; opterr = 0;

	if ( rrd_fetch(argv->count-1, &argv->args[1], &start,&end,&step,&ds_cnt,&ds_namv,&return_data) != -1 ) {
		array_init(return_value);
		add_assoc_long(return_value, "start", start);
		add_assoc_long(return_value, "end", end);
		add_assoc_long(return_value, "step", step);
		add_assoc_long(return_value, "ds_cnt", ds_cnt);

		MAKE_STD_ZVAL(p_ds_namv);
		array_init(p_ds_namv);

		if (ds_namv)
		{
			for (i = 0; i < ds_cnt; i++)
			{
				add_next_index_string(p_ds_namv, ds_namv[i], 1);
				free(ds_namv[i]);
			}
			free(ds_namv);
		}

		add_assoc_zval(return_value,"ds_namv",p_ds_namv);

		MAKE_STD_ZVAL(p_data);
		array_init(p_data);

		if (return_data) {
			datap = return_data;

			for (i = start; i <= end; i += step)
				for (j = 0; j < ds_cnt; j++)
					add_next_index_double(p_data, *(datap++));

			free(return_data);
		}

		add_assoc_zval(return_value,"data",p_data);
	} else {
		RETVAL_FALSE;
	}

	__rrd_argument_free(argv);

	return;
}
/* }}} */

/* {{{ proto string rrd_error(void)
	Get the error message set by the last rrd tool function call */
PHP_FUNCTION(rrd_error)
{
	char *msg;

	if ( rrd_test_error() )
	{
		msg = rrd_get_error();

		RETVAL_STRING(msg, 1);
		rrd_clear_error();
	}

	return;
}
/* }}} */

/* {{{ proto void rrd_clear_error(void)
	Clear the error set by the last rrd tool function call */
PHP_FUNCTION(rrd_clear_error)
{
	if ( rrd_test_error() )
		rrd_clear_error();

	return;
}
/* }}} */

/* {{{ proto int rrd_update(string file, string opt)
	Update an RRD file with values specified */
PHP_FUNCTION(rrd_update)
{
	char *file;
	int file_len;
	zval *opt;

	HashTable *args_arr;

	if ( rrd_test_error() )
		rrd_clear_error();

	if ( zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sz", &file, &file_len, &opt) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(file TSRMLS_CC))
		RETURN_FALSE;

	rrd_arg_array *argv;

	if (Z_TYPE_P(opt) == IS_STRING) {
		argv = (rrd_arg_array *)emalloc(sizeof(rrd_arg_array));
		if(!argv)
			RETURN_FALSE;

		argv->count = 4;
		argv->args = (char **)emalloc(argv->count * sizeof(char *));
		if(!argv->args) {
			efree(argv);
			RETURN_FALSE;
		}

		argv->args[0] = "dummy";
		argv->args[1] = estrdup("update");
		argv->args[2] = estrdup(file);
		argv->args[3] = estrdup(Z_STRVAL_P(opt));
	} else if (Z_TYPE_P(opt) == IS_ARRAY) {
		args_arr = Z_ARRVAL_P(opt);

		rrd_arg_array *argv = __rrd_argument_strdup("update", file, args_arr);
		if(!argv)
			RETURN_FALSE;
	} else {
		RETURN_FALSE;
	}

	optind = 0; opterr = 0;

	if ( rrd_update(argv->count - 1, &argv->args[1]) != -1 )
	{
		RETVAL_TRUE;
	}
	else
	{
		RETVAL_FALSE;
	}

	__rrd_argument_free(argv);

	return;
}
/* }}} */

/* {{{ proto int rrd_last(string file)
	Gets last update time of an RRD file */
PHP_FUNCTION(rrd_last)
{
	pval *file;
	unsigned long retval;

	if ( rrd_test_error() )
		rrd_clear_error();

	if (zend_get_parameters(ht, 1, &file) == SUCCESS)
	{
		convert_to_string(file);

		//BenBE: Added Basedir Restriction ...
		if (php_check_open_basedir(file->value.str.val TSRMLS_CC)) {
			RETURN_FALSE;
		}

		retval = rrd_last_r(file->value.str.val);

		RETVAL_LONG(retval);
	}
	else
	{
		WRONG_PARAM_COUNT;
	}

	return;
}
/* }}} */

/* {{{ proto int rrd_lastupdate(string file)
	Gets last update details of an RRD file */
PHP_FUNCTION(rrd_lastupdate)
{
	char *file;
	int file_len;
	unsigned long retval;

	time_t last_update;
	unsigned long ds_cnt;
	char **ds_namv;
	char **last_ds;

	int i;

	zval *p_ds_namv, *p_data;

	if ( rrd_test_error() )
		rrd_clear_error();

	if ( zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &file, &file_len) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(file TSRMLS_CC))
		RETURN_FALSE;

	if (rrd_lastupdate_r(file, &last_update, &ds_cnt, &ds_namv, &last_ds) == -1)
		RETURN_FALSE;

	array_init(return_value);
	add_assoc_long(return_value, "last_update", last_update);
	add_assoc_long(return_value, "ds_cnt", ds_cnt);

	MAKE_STD_ZVAL(p_ds_namv);
	array_init(p_ds_namv);

	if (ds_namv) {
		for (i = 0; i < ds_cnt; i++) {
			add_next_index_string(p_ds_namv, ds_namv[i], 1);
			free(ds_namv[i]);
		}
		free(ds_namv);
	}

	add_assoc_zval(return_value,"ds_namv",p_ds_namv);

	MAKE_STD_ZVAL(p_data);
	array_init(p_data);

	if (last_ds) {
		for (i = 0; i < ds_cnt; i++) {
			add_next_index_string(p_data, last_ds[i], 1);
			free(last_ds[i]);
		}
		free(last_ds);
	}

	add_assoc_zval(return_value,"last_ds",p_data);

	return;
}
/* }}} */

/* {{{ proto int rrd_first(string file, int rraindex)
	Gets first update time of an RRD file */
PHP_FUNCTION(rrd_first)
{
	char *file;
	int file_len;
	long rraindex = 0;
	unsigned long retval;

	if ( rrd_test_error() )
		rrd_clear_error();

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l", &file, &file_len, &rraindex) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(file))
		RETURN_FALSE;

	retval = rrd_first_r(file, rraindex);

	RETVAL_LONG(retval);
	return;
}
/* }}} */

/* {{{ proto int rrd_info(string file)
	Gets last update details of an RRD file */

void rrd_info_add_assoc(
	zval *array,
	char *key,
	rrd_info_t *data);

void rrd_info_add_assoc(
	zval *array,
	char *key,
	rrd_info_t *data) {

	switch (data->type) {
		case RD_I_VAL:
			if (isnan(data->value.u_val))
				add_assoc_string(array,key,"NaN",1);
			else
				add_assoc_double(array,key,data->value.u_val);
			break;
		case RD_I_CNT:
			add_assoc_long(array,key,data->value.u_cnt);
			break;
		case RD_I_INT:
			add_assoc_long(array,key,data->value.u_int);
			break;
		case RD_I_STR:
			add_assoc_string(array,key,data->value.u_str,1);
			free(data->value.u_str);
			break;
		case RD_I_BLO:
			// ignore
			free(data->value.u_blo.ptr);
			break;
	}
}

PHP_FUNCTION(rrd_info)
{
	char *file;
	int file_len;

	char tmpstr1[200];
	char tmpstr2[200];
	char tmpstr3[200];
	char tmpstr4[200];

	char tmpkey1[200] = "";
	char tmpkey2[200] = "";
	char tmpkey3[200] = "";

	rrd_info_t *data, *save;

	zval *p_ds, *p_rra, *p_cdp_prep, *temp1, *temp2, *temp3, *temp4;

	if ( rrd_test_error() )
		rrd_clear_error();

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &file, &file_len) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(file)) {
		RETURN_FALSE;
	}

	char *argv[3];
	argv[0] = "dummy";
	argv[1] = estrdup("info");
	argv[2] = estrndup(file,file_len);

	optind = 0; opterr = 0;

	data = rrd_info(2, &argv[1]);

	efree(argv[2]);
	efree(argv[1]);

	if (NULL == data)
		RETURN_FALSE;

	array_init(return_value);

	MAKE_STD_ZVAL(p_ds);
	array_init(p_ds);

	MAKE_STD_ZVAL(p_rra);
	array_init(p_rra);

	MAKE_STD_ZVAL(temp1);
	array_init(temp1);

	MAKE_STD_ZVAL(temp2);
	array_init(temp2);

	MAKE_STD_ZVAL(temp3);
	array_init(temp3);

	MAKE_STD_ZVAL(temp4);
	array_init(temp4);

	while (data) {
		save = data;
		if (strncmp(data->key,"ds",2) == 0 && sscanf(data->key,"ds[%100[a-zA-Z0-9]].%100s",&tmpstr1,&tmpstr2) == 2) {
			if (strcmp(tmpkey1,tmpstr1) != 0) {
				if (strlen(tmpkey1) > 0) {
					add_assoc_zval(p_ds,tmpkey1,temp1);

					MAKE_STD_ZVAL(temp1);
					array_init(temp1);
				}
				strcpy(tmpkey1,tmpstr1);
			}

			rrd_info_add_assoc(temp1,tmpstr2,data);
		} else if (strncmp(data->key,"rra",3) == 0 && sscanf(data->key,"rra[%4[0-9]].%100s",&tmpstr1,&tmpstr2) == 2) {
			if (strcmp(tmpkey2,tmpstr1) != 0) {
				if (strlen(tmpkey2) > 0) {
					if (strlen(tmpkey3) > 0) {
						add_assoc_zval(temp3,tmpkey3,temp4);
						add_assoc_zval(temp2,"cdp_prep",temp3);
					}

					add_assoc_zval(p_rra,tmpkey2,temp2);

					MAKE_STD_ZVAL(temp2);
					array_init(temp2);

					MAKE_STD_ZVAL(temp3);
					array_init(temp3);

					MAKE_STD_ZVAL(temp4);
					array_init(temp4);
				}
				strcpy(tmpkey2,tmpstr1);
				strcpy(tmpkey3,"");
			}

			if (strncmp(tmpstr2,"cdp_prep",8) == 0 && sscanf(tmpstr2,"cdp_prep[%4[0-9]].%100s",&tmpstr3,&tmpstr4) == 2) {
				if (strcmp(tmpkey3,tmpstr3) != 0) {
					if (strlen(tmpkey3) > 0) {
						add_assoc_zval(temp3,tmpkey3,temp4);

						MAKE_STD_ZVAL(temp4);
						array_init(temp4);
					}
					strcpy(tmpkey3,tmpstr3);
				}

				rrd_info_add_assoc(temp4,tmpstr4,data);
			} else {
				rrd_info_add_assoc(temp2,tmpstr2,data);
			}
		} else {
			rrd_info_add_assoc(return_value,data->key,data);
		}

		free(data->key);

		data = data->next;
		free(save);
	}

	if (strlen(tmpkey1) > 0) {
		add_assoc_zval(p_ds,tmpkey1,temp1);
	}

	if (strlen(tmpkey2) > 0) {
		if (strlen(tmpkey3) > 0) {
			add_assoc_zval(temp3,tmpkey3,temp4);
			add_assoc_zval(temp2,"cdp_prep",temp3);
		}

		add_assoc_zval(p_rra,tmpkey2,temp2);
	}

	add_assoc_zval(return_value,"ds",p_ds);
	add_assoc_zval(return_value,"rra",p_rra);

	return;
}
/* }}} */

/* {{{ proto string rrd_dump(string file, [int header])
	Gets an XML dump of an RRD file */

typedef struct rrd_dump_memout {
	void *buf;
	void *cur;
	size_t len;
	} rrd_dump_memout_t;

size_t rrd_dump_opt_cb_memout(
	const void *data,
	size_t len,
	void *user);

size_t rrd_dump_opt_cb_memout(
	const void *data,
	size_t len,
	void *user) {

	if(!user) {
		return 0;
	}

	rrd_dump_memout_t *userdata = (rrd_dump_memout_t *)user;

	//Validate the pointers
	if(!userdata->buf) {
		return 0;
	}
	if(userdata->buf > userdata->cur) {
		userdata->cur = userdata->buf + userdata->len + 1; //Enforce validation
	}
	if(userdata->buf + userdata->len < userdata->cur) {
		userdata->cur = userdata->buf + strnlen(userdata->buf, userdata->len);
	}

	//Check if we have enough memory left
	if(userdata->buf + userdata->len <= userdata->cur + len + 1) {
		size_t buflen =userdata->cur - userdata->buf;
		size_t tmp = buflen + len + 1; //Get the memory size we need
		tmp += 16384;   //Ask for some more
		tmp &= ~8191;   //But round down to an 8KB block
		userdata->buf = erealloc(userdata->buf, tmp);
		if(!userdata->buf) {
			userdata->buf = NULL;
			userdata->cur = NULL;
			userdata->len = 0;
			return 0;
		}
		userdata->cur = userdata->buf + buflen;
		userdata->len = tmp;
	}

	strncpy(userdata->cur, data, len);
	userdata->cur += len;
	*(char *)userdata->cur = 0;

	return len;
}

PHP_FUNCTION(rrd_dump)
{
#if HAVE_RRD_12X
	char *file;
	int file_len;
	long header = RRD_DUMP_XML_NONE;

	if ( rrd_test_error() )
		rrd_clear_error();

	if (!ZEND_NUM_ARGS() || (ZEND_NUM_ARGS() > 2))
		WRONG_PARAM_COUNT;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l", &file, &file_len, &header) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(file TSRMLS_CC))
		RETURN_FALSE;

	rrd_dump_memout_t rrd_dump_data;
	rrd_dump_data.cur = rrd_dump_data.buf = emalloc(rrd_dump_data.len = 8192);
	if(!rrd_dump_data.buf || !rrd_dump_data.cur)
	{
		RETURN_FALSE;
	}
	*(char *)rrd_dump_data.cur = 0;

	rrd_dump_cb_r(file, header, rrd_dump_opt_cb_memout, (void *)&rrd_dump_data);

	RETURN_STRINGL(rrd_dump_data.buf, rrd_dump_data.cur - rrd_dump_data.buf, 0);
#else
	RETURN FALSE;
#endif
}
/* }}} */

/* {{{ proto array rrd_restore(string infile, string outfile)
	Restores an RRD file from a XML dump */
PHP_FUNCTION(rrd_restore)
{
	char *infile, *outfile;
	int infile_len, outfile_len;

	if ( rrd_test_error() )
		rrd_clear_error();

	if (ZEND_NUM_ARGS() != 2)
		WRONG_PARAM_COUNT;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss", &infile, &infile_len, &outfile, &outfile_len) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(infile TSRMLS_CC))
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(outfile TSRMLS_CC))
		RETURN_FALSE;

	char *argv[4];
	argv[0] = "dummy";
	argv[1] = estrdup("restore");
	argv[2] = estrdup(infile);
	argv[3] = estrdup(outfile);

	if ( rrd_restore(3, &argv[1]) != -1 )
	{
		RETVAL_TRUE;
	}
	else
	{
		RETVAL_FALSE;
	}

	efree(argv[3]);
	efree(argv[2]);
	efree(argv[1]);

	return;
}
/* }}} */

/* {{{ proto int rrd_create(string file, array args_arr, [int argc])
	Create an RRD file with the options passed (passed via array) */
PHP_FUNCTION(rrd_create)
{
	char *file;
	int file_len;
	zval *args;
	long dummy;

	HashTable *args_arr;

	if ( rrd_test_error() )
		rrd_clear_error();

	if ( zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sa|l", &file, &file_len, &args, &dummy) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(file TSRMLS_CC))
		RETURN_FALSE;

	args_arr = Z_ARRVAL_P(args);

	rrd_arg_array *argv = __rrd_argument_strdup("create", file, args_arr);
	if(!argv)
		RETURN_FALSE;

	optind = 0;  opterr = 0;

	if ( rrd_create(argv->count - 1, &argv->args[1]) != -1 )
	{
		RETVAL_TRUE;
	}
	else
	{
		RETVAL_FALSE;
	}

	__rrd_argument_free(argv);

	return;
}
/* }}} */

/* {{{ proto int rrd_tune(string file, array args_arr, [int argc])
	Tune an RRD file with the options passed (passed via array) */
PHP_FUNCTION(rrd_tune)
{
	char *file;
	int file_len;
	zval *args;
	long dummy;

	HashTable *args_arr;

	if ( rrd_test_error() )
		rrd_clear_error();

	if (ZEND_NUM_ARGS() < 2)
		WRONG_PARAM_COUNT;

	if ( zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sa", &file, &file_len, &args) == FAILURE )
		RETURN_FALSE;

	//BenBE: Added Basedir Restriction ...
	if (php_check_open_basedir(file TSRMLS_CC))
		RETURN_FALSE;

	args_arr = Z_ARRVAL_P(args);

	rrd_arg_array *argv = __rrd_argument_strdup("tune", file, args_arr);
	if(!argv)
		RETURN_FALSE;

	optind = 0;  opterr = 0;

	if ( rrd_tune(argv->count - 1, &argv->args[1]) != -1 )
	{
		RETVAL_TRUE;
	}
	else
	{
		RETVAL_FALSE;
	}

	__rrd_argument_free(argv);

	return;
}
/* }}} */

PHP_FUNCTION(rrdtool_info)
{
	pval *as_array;
	zend_bool b_as_array = 0;

	if (ZEND_NUM_ARGS() > 1)
		WRONG_PARAM_COUNT;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|b", &b_as_array) == FAILURE)
		b_as_array = 0;

	if(b_as_array) {
		array_init(return_value);

		add_assoc_string(return_value, "Extension Name", "rrdtool", 1);
		add_assoc_string(return_value, "Extension Version", PHP_RRD_VERSION_STRING, 1);
#ifdef COMPILE_DL_RRDTOOL
		add_assoc_string(return_value, "Extension Linkage", "Extension", 1);
#else
		add_assoc_string(return_value, "Extension Linkage", "Bundled", 1);
#endif
		add_assoc_string(return_value, "Extension Copyright",
			"Portions copyright 2005 by Dale Walsh (buildsmart@daleenterprise.com), 2009 by Dan Cech, 2009 by Benny Baumann\n"
			"Portions relating to rrdtool 1999, 2000, 2001, 2002, 2003, 2004, 2005 by Tobias Oetiker.", 1);
		add_assoc_string(return_value, "Extension License", "PHP License", 1);

		add_assoc_string(return_value, "Build System", PHP_UNAME, 1);
		add_assoc_string(return_value, "Build Timestamp", __DATE__ " " __TIME__, 1);
		add_assoc_string(return_value, "Support", "Enabled", 1);

		add_assoc_string(return_value, "RRDTool Library Name", "librrd", 1);
		add_assoc_string(return_value, "RRDTool Library Version", rrd_strversion(), 1);
		add_assoc_string(return_value, "RRDTool Library License",
			"<b>Permission has been granted to copy, distribute and modify "
			"rrd in any context without fee, including a commercial application, "
			"provided that this notice is present in user-accessible supporting "
			"documentation.</b>\n\n"

			"This does not affect your ownership of the derived work itself, "
			"and the intent is to assure proper credit for the authors of rrdtool, "
			"not to interfere with your productive use of rrdtool. "
			"If you have questions, ask. "
			"\"Derived works\" includes all programs that utilize the library. "
			"Credit must be given in user-accessible documentation.\n\n"

			"<b>This software is provided \"AS IS\"</b>. The copyright holders "
			"disclaim all warranties, either express or implied, including but "
			"not limited to implied warranties of merchantability and fitness "
			"for a particular purpose, with respect to this code and "
			"accompanying documentation.\n\n"
			, 1);

		add_assoc_string(return_value, "Special Thanks",
			"Perl by Larry Wall\n"
			"gd library by Thomas Boutell\n"
			"gifcode from David Koblas\n"
			"libpng by Glenn Randers-Pehrson / Andreas Eric Dilger / Guy Eric Schalnat\n"
			"cgilib by Martin Schulze\n"
			"zlib by Jean-loup Gailly and Mark Adler\n"
			"php4 and php5 bindings by Dale Walsh, Dan Cech and Benny Baumann", 1);
	} else {
		PUTS("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"DTD/xhtml1-transitional.dtd\">\n");
		PUTS("<html>\n");
		PUTS("<head>\n");
		PUTS("<style type=\"text/css\"><!--\n");
		PUTS("body {background-color: #ffffff; color: #000000;}\n");
		PUTS("body, td, th, h1, h2 {font-family: sans-serif;}\n");
		PUTS("pre {margin: 0px; font-family: monospace;}\n");
		PUTS("a:link {color: #000099; text-decoration: none; background-color: #ffffff;}\n");
		PUTS("a:hover {text-decoration: underline;}\n");
		PUTS("table {border-collapse: collapse;}\n");
		PUTS(".center {text-align: center;}\n");
		PUTS(".center table { margin-left: auto; margin-right: auto; text-align: left;}\n");
		PUTS(".center th { text-align: center !important; }\n");
		PUTS("td, th { border: 1px solid #000000; font-size: 75%; vertical-align: baseline;}\n");
		PUTS("h1 {font-size: 150%;}\n");
		PUTS("h2 {font-size: 125%;}\n");
		PUTS(".p {text-align: left;}\n");
		PUTS(".e {background-color: #ccccff; font-weight: bold; color: #000000;}\n");
		PUTS(".h {background-color: #9999cc; font-weight: bold; color: #000000;}\n");
		PUTS(".v {background-color: #cccccc; color: #000000;}\n");
		PUTS("i {color: #666666; background-color: #cccccc;}\n");
		PUTS("img {float: right; border: 0px;}\n");
		PUTS("hr {width: 600px; background-color: #cccccc; border: 0px; height: 1px; color: #000000;}\n");
		PUTS("//--></style>\n");
		PUTS("<title>rrdtool_info()</title>\n");
		PUTS("</head>\n");
		PUTS("<body>\n");
		PUTS("<div class=\"center\">\n");

		php_info_print_box_start(1);
		PUTS("<a href=\"http://oss.oetiker.ch/rrdtool/\" target=\"rrdtool\"><img border=\"0\" src=\"");
		if (SG(request_info).request_uri) {
			PUTS(SG(request_info).request_uri);
		}
		PUTS("?="RRDTOOL_LOGO_GUID"\" alt=\"RRDTool logo\" /></a>\n");
		php_printf("<h1 class=\"p\">rrdtool Version %s (%s)</h1>\n", PHP_RRD_VERSION_STRING, rrd_strversion());
		php_info_print_box_end();
		php_info_print_table_start();
		php_info_print_table_row(2, "System", PHP_UNAME );
		php_info_print_table_row(2, "Build Date", __DATE__ " " __TIME__ );
		php_info_print_table_row(2, "rrdtool Support","Enabled");
		php_info_print_table_end();

		PUTS("<h2>RRDTOOL Copyright</h2>\n");
		php_info_print_box_start(0);
		PUTS("COPYRIGHT STATEMENT FOLLOWS THIS LINE</p>\n<blockquote>\n");
		PUTS("<p>Portions copyright 2005 by Dale Walsh (buildsmart@daleenterprise.com).</p>\n");
		PUTS("<p>Portions relating to rrdtool 1999, 2000, 2001, 2002, 2003, 2004, 2005 by Tobias Oetiker.</p>\n");
		php_info_print_box_end();
		PUTS("<h2>RRDTOOL License</h2>\n");
		php_info_print_box_start(0);
		PUTS("<p><b>Permission has been granted to copy, distribute and modify rrd in any context without fee, including a commercial application, provided that this notice is present in user-accessible supporting documentation. </b></p>");
		PUTS("<p>This does not affect your ownership of the derived work itself, and the intent is to assure proper credit for the authors of rrdtool, not to interfere with your productive use of rrdtool. If you have questions, ask. \"Derived works\" ");
		PUTS("includes all programs that utilize the library. Credit must be given in user-accessible documentation.</p>\n");
		PUTS("<p><b>This software is provided \"AS IS.\"</b> The copyright holders disclaim all warranties, either express or implied, including but not limited to implied warranties of merchantability and fitness for a particular purpose, ");
		PUTS("with respect to this code and accompanying documentation.</p>\n");
		php_info_print_box_end();
		PUTS("<h2>Special Thanks</h2>\n");
		php_info_print_box_start(0);
		PUTS("<p>Perl by Larry Wall");
		PUTS("<p>gd library by Thomas Boutell");
		PUTS("<p>gifcode from David Koblas");
		PUTS("<p>libpng by Glenn Randers-Pehrson / Andreas Eric Dilger / Guy Eric Schalnat");
		PUTS("<p>cgilib by Martin Schulze");
		PUTS("<p>zlib by Jean-loup Gailly and Mark Adler");
		PUTS("<p>Portions relating to php4 and php5 bindings, Dale Walsh (buildsmart@daleenterprise.com)");
		php_info_print_box_end();

		PUTS("</div>\n");
		PUTS("</body>\n");
		PUTS("</html>\n");
	}
}
/* }}} */

PHP_FUNCTION(rrdtool_logo_guid)
{
	if (ZEND_NUM_ARGS() != 0)
		WRONG_PARAM_COUNT;

	RETURN_STRINGL(RRDTOOL_LOGO_GUID, sizeof(RRDTOOL_LOGO_GUID)-1, 1);
}
/* }}} */

#endif	/* HAVE_RRDTOOL */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
